from __future__ import annotations

import asyncio
import random
import re
from contextlib import asynccontextmanager
import string
from typing import Awaitable, Callable, Iterable, List, Optional, Sized, Union

from loguru import logger
from pyrogram import filters
from pyrogram.enums import ChatType
from pyrogram.errors import UsernameNotOccupied, UserNotParticipant, FloodWait, ChannelInvalid, ChannelPrivate
from pyrogram.handlers import EditedMessageHandler, MessageHandler
from pyrogram.types import Message, User

from embykeeper import __name__ as __product__
from embykeeper.utils import show_exception, to_iterable, truncate_str, AsyncCountPool, optional
from embykeeper.config import config
from embykeeper.runinfo import RunContext, RunStatus

from ..pyrogram import Client
from ..link import Link

__ignore__ = True


class Session:
    """回复检测会话, 用于检测跟随回复."""

    def __init__(self, reply, follows=None, delays=None):
        self.reply = reply
        self.follows = follows
        self.delays = delays
        self.lock = asyncio.Lock()
        self.delayed = asyncio.Event()
        self.followed = asyncio.Event()
        self.canceled = asyncio.Event()
        if not self.follows:
            return self.followed.set()

    async def delay(self):
        if not self.delayed:
            return self.delayed.set()
        if isinstance(self.delays, Sized) and len(self.delays) == 2:
            time = random.uniform(*self.delays)
        else:
            time = self.delays
        await asyncio.sleep(time)
        self.delayed.set()

    async def follow(self):
        async with self.lock:
            self.follows -= 1
            if self.follows <= 0:
                self.followed.set()
            return self.follows

    async def wait(self, timeout=240):
        task = asyncio.create_task(self.delay())
        try:
            await asyncio.wait_for(asyncio.gather(self.delayed.wait(), self.followed.wait()), timeout=timeout)
        except asyncio.TimeoutError:
            return False
        else:
            return not self.canceled.is_set()

    async def cancel(self):
        self.canceled.set()
        self.delayed.set()
        self.followed.set()


class UniqueUsername(dict):
    """独特名称类, 用于抢注."""

    def __getitem__(self, user: User):
        if not user.id in self:
            self[user.id] = self.get_unique(user)
        return dict.__getitem__(self, user.id)

    @staticmethod
    def get_unique(user: User):
        """获得一个独特名称, 该名称将在程序运行全周期一致."""
        log = logger.bind(scheme="telemonitor", username=user.name)
        if user.username:
            unique = user.username
        else:
            unique: str = user.name.lower()
        unique = re.sub(r"[^A-Za-z0-9]", "", unique)
        random_bits = 10 - len(unique)
        if random_bits:
            random_bits = "".join(random.choice(string.digits) for _ in range(random_bits))
            unique = unique + random_bits
        log.info(
            f'([magenta]默认[/]) 当监控到开注时, 将以用户名 "{unique}" 注册, 请[yellow]保证[/]具有一定独特性以避免注册失败.'
        )
        return unique


class Monitor:
    """监控器类, 可以检测某个人在某个群中发送了某种模式的信息, 并触发特定的动作 (回复/向机器人注册) 等, 用于答题/抢注等."""

    group_pool = AsyncCountPool(base=1000)
    unique_cache = UniqueUsername()

    name: str = None  # 监控器名称
    chat_name: Union[str, int] = None  # 监控的群聊名称
    chat_allow_outgoing: bool = False  # 是否支持自己发言触发
    chat_user: Union[str, List[str]] = []  # 仅被列表中用户的发言触发 (支持 username / userid)
    chat_keyword: Union[str, List[str]] = []  # 仅当消息含有列表中的关键词时触发, 支持 regex
    chat_except_keyword: Union[str, List[str]] = []  # 消息含有列表中的关键词时不触发, 支持 regex
    chat_probability: float = 1.0  # 发信概率 (0最低, 1最高)
    chat_delay: int = 0  # 发信延迟 (s)
    chat_follow_user: int = 0  # 需要等待 N 个用户发送 {chat_reply} 方可回复
    chat_reply: Union[
        str, Callable[[Message, Optional[Union[str, List[str]]]], Union[str, Awaitable[str]]]
    ] = None  # 回复的内容, 可以为恒定字符串或函数或异步函数
    notify_create_name: bool = False  # 启动时生成 unique name 并提示, 用于抢注
    allow_edit: bool = False  # 编辑消息内容后也触发
    trigger_interval: float = 2  # 每次触发的最低时间间隔
    trigger_sim: int = 1  # 同时触发的最大并行数
    trigger_max_time: float = 120  # 触发后处理的最长时间
    additional_auth: List[str] = []  # 额外认证要求
    debug_no_log = False  # 调试模式不显示冗余日志
    allow_caption: bool = True  # 是否允许带照片的消息
    allow_text: bool = True  # 是否允许不带照片的消息
    init_first: bool = False  # 先执行自定义初始化函数, 再进行加入群组分析

    def __init__(
        self,
        client: Client,
        context: RunContext = None,
        config: dict = {},
    ):
        """
        监控器类.
        参数:
            client: Pyrogram 客户端
            context: 运行时上下文
            nofail: 启用错误处理外壳, 当错误时报错但不退出
            basedir: 文件存储默认位置
            proxy: 代理配置
            config: 当前监控器的特定配置
        """
        self.client = client
        self.ctx = context or RunContext.prepare()

        self.config = config
        self.log = logger.bind(scheme="telemonitor", name=self.name, username=client.me.name)
        self.session = None
        self.failed = asyncio.Event()
        if self.trigger_sim:
            self.sem = asyncio.Semaphore(self.trigger_sim)
        else:
            self.sem = None

    def get_filter(self):
        """设定要监控的目标."""
        filter = filters.all
        if self.chat_name:
            filter = filter & filters.chat(self.chat_name)
        if not self.chat_allow_outgoing:
            filter = filter & (~filters.outgoing)
        return filter

    def get_handlers(self):
        """设定要监控的更新的类型."""
        handlers = [MessageHandler(self._message_handler, self.get_filter())]
        if self.allow_edit:
            handlers.append(EditedMessageHandler(self._message_handler, self.get_filter()))
        return handlers

    @asynccontextmanager
    async def listener(self):
        """执行监控上下文."""
        group = await self.group_pool.append(self)
        handlers = self.get_handlers()
        for h in handlers:
            await self.client.add_handler(h, group=group)
        yield
        for h in handlers:
            try:
                await self.client.remove_handler(h, group=group)
            except ValueError:
                pass

    async def _start(self):
        """监控器的入口函数的错误处理外壳."""
        try:
            return await self.start()
        except Exception as e:
            if config.nofail:
                self.log.warning(f"发生初始化错误, 监控停止.")
                show_exception(e, regular=False)
                return self.ctx.finish(RunStatus.FAIL, "异常错误")
            else:
                raise

    async def start(self):
        """监控器的入口函数."""
        self.ctx.start(RunStatus.INITIALIZING)
        if self.init_first:
            if not await self.init():
                self.log.bind(log=True).warning(f"机器人状态初始化失败, 监控将停止.")
                return self.ctx.finish(RunStatus.FAIL, "初始化失败")
        chat = None
        if self.chat_name:
            try:
                chat = await self.client.get_chat(self.chat_name)
            except UsernameNotOccupied:
                self.log.warning(f'初始化错误: 群组 "{self.chat_name}" 不存在.')
                return self.ctx.finish(RunStatus.IGNORE, "群组不存在")
            except UserNotParticipant:
                self.log.info(f'跳过监控: 尚未加入群组 "{chat.title}".')
                return self.ctx.finish(RunStatus.IGNORE, "未加入群组")
            except KeyError as e:
                self.log.info(f"初始化错误: 无法访问, 您可能已被封禁.")
                show_exception(e)
                return self.ctx.finish(RunStatus.FAIL, "无法访问群组")
            except (ChannelInvalid, ChannelPrivate) as e:
                self.log.info(f"跳过监控: 私有群组, 未加入, 已跳过.")
                return self.ctx.finish(RunStatus.IGNORE, "未加入群组")
            except FloodWait as e:
                self.log.info(f"初始化信息: Telegram 要求等待 {e.value} 秒.")
                if e.value < 360:
                    await asyncio.sleep(e.value)
                else:
                    self.log.info(
                        f"初始化信息: Telegram 要求等待 {e.value} 秒, 您可能操作过于频繁, 监控器将停止."
                    )
                    return self.ctx.finish(RunStatus.FAIL, "操作过于频繁")
            try:
                if chat.type in (ChatType.GROUP, ChatType.SUPERGROUP):
                    await chat.get_member("me")
            except UserNotParticipant:
                self.log.info(f'跳过监控: 尚未加入群组 "{chat.title}".')
                return self.ctx.finish(RunStatus.IGNORE, "未加入群组")

        if self.additional_auth:
            for a in self.additional_auth:
                if not await Link(self.client).auth(a, log_func=self.log.info):
                    return self.ctx.finish(RunStatus.IGNORE, "需要额外认证")

        if not self.init_first:
            if not await self.init():
                self.log.bind(log=True).warning(f"机器人状态初始化失败, 监控将停止.")
                return self.ctx.finish(RunStatus.FAIL, "初始化失败")

        if self.notify_create_name:
            self.unique_name = self.get_unique_name()

        if chat:
            if chat.username:
                spec = (
                    f"[green]{chat.title}[/] [gray50](@{chat.username})[/]"
                    if chat.title
                    else f"@{chat.username}"
                )
            else:
                spec = f"[green]{chat.title}[/]" if chat.title else f"[green]{chat.id}[/]"
        else:
            spec = "关键词: " + truncate_str(",".join(to_iterable(self.chat_keyword)), 60)

        self.log.info(f"开始监视: {spec}.")
        self.ctx.status = RunStatus.RUNNING

        async with self.listener():
            await self.failed.wait()
            self.log.error(f"发生错误, 不再监视: {spec}.")
            return self.ctx.finish(RunStatus.ERROR, "异常错误")

    async def init(self):
        """可重写的初始化函数, 在读取聊天后运行, 在执行监控前运行, 返回 False 将视为初始化错误."""
        return True

    def keys(self, message: Message):
        """提取信息中的 keys, 注意该函数可能通过类直接调用 (self 为 cls)."""
        sender = message.from_user
        if (
            sender
            and self.chat_user
            and not any(i in to_iterable(self.chat_user) for i in (sender.id, sender.username))
        ):
            return
        text = message.text or message.caption
        if text and self.chat_except_keyword:
            for k in to_iterable(self.chat_except_keyword):
                if re.search(k, text, re.IGNORECASE):
                    return
        if self.chat_keyword:
            for k in to_iterable(self.chat_keyword):
                if k is None or text is None:
                    if k is None and text is None:
                        yield None
                else:
                    for m in re.findall(k, text, re.IGNORECASE):
                        yield m
        else:
            yield text

    async def get_reply(self, message: Message, key: Union[str, List[str]]):
        """根据 keys 生成回复内容."""
        if callable(self.chat_reply):
            result = self.chat_reply(message, key)
            if asyncio.iscoroutinefunction(self.chat_reply):
                return await result
            else:
                return result
        else:
            reply = self.chat_reply
            if isinstance(reply, str):
                if isinstance(key, str):
                    # If key is a single value, only replace $1
                    reply = re.sub(r"(?<!\\)\$1", str(key), reply)
                elif key is not None:
                    # Replace $1, $2, etc. with corresponding values from key list
                    for i, k in enumerate(key, 1):
                        reply = re.sub(r"(?<!\\)\$" + str(i), str(k), reply)
            return reply

    @staticmethod
    def get_spec(keys):
        """返回 keys 的简要表示."""
        if keys is None:
            return "<仅媒体消息>"
        if isinstance(keys, Iterable) and not isinstance(keys, str):
            keys = " ".join([str(k).strip() for k in keys])
        return truncate_str(keys.replace("\n", " ").strip(), 30)

    async def _message_handler(self, client: Client, message: Message):
        """消息处理入口函数的错误处理外壳."""
        try:
            await self.message_handler(client, message)
        except OSError as e:
            self.log.info(f'发生错误: "{e}", 忽略.')
        except Exception as e:
            self.failed.set()
            if config.nofail:
                self.log.warning(f"发生错误, 监控器将停止.")
                show_exception(e, regular=False)
            else:
                raise
        finally:
            message.continue_propagation()

    async def message_handler(self, client: Client, message: Message):
        """消息处理入口函数, 控制是否回复以及等待回复."""
        if message.caption and not self.allow_caption:
            return
        if message.text and not self.allow_text:
            return
        for key in self.keys(message):
            spec = self.get_spec(key)
            if not self.debug_no_log:
                if spec:
                    self.log.info(f"监听到关键信息: {truncate_str(spec, 30)}")
                else:
                    self.log.info(f"监听到关键信息.")
            if random.random() >= self.chat_probability:
                self.log.info(f"由于概率设置, 不予回应: {truncate_str(spec, 30)}")
                return
            reply = await self.get_reply(message, key)
            if self.session:
                await self.session.cancel()
            if self.chat_follow_user:
                self.log.info(f"将等待{self.chat_follow_user}个人回复: {reply}")
            self.session = Session(reply, follows=self.chat_follow_user, delays=self.chat_delay)
            if await self.session.wait():
                self.session = None
                async with optional(self.sem):
                    try:
                        await asyncio.wait_for(self.on_trigger(message, key, reply), self.trigger_max_time)
                    except asyncio.TimeoutError:
                        self.log.warning(f"处理超时: {truncate_str(spec, 30)}")
                    await asyncio.sleep(self.trigger_interval)
        else:
            if self.session and not self.session.followed.is_set():
                text = message.text or message.caption
                if self.session.reply == text:
                    now = await self.session.follow()
                    self.log.info(
                        f'从众计数 ({self.chat_follow_user - now}/{self.chat_follow_user}): "{message.from_user.name}"'
                    )

    async def on_trigger(self, message: Message, key: Optional[Union[List[str], str]], reply: str):
        """
        可修改的回调函数.
        参数:
            message: 引发回复的消息
            key:
                当 chat_keyword 没有 capturing groups 时, 类型为 str, 内容为 regex 的 match
                当 chat_keyword 仅有一个 capturing groups 时, 类型为 str, 内容为 regex 的唯一一个 capturing groups 的对应值
                当 chat_keyword 有多个 capturing groups 时, 类型为 list(str), 内容为 regex 的各个 capturing groups 的对应值
        """
        if reply:
            return await self.client.send_message(message.chat.id, reply)

    def get_unique_name(self):
        """获取唯一性用户名, 用于注册."""
        unique_name = self.config.get("unique_name", None)
        if unique_name:
            self.log.info(f'根据您的设置, 当监控到开注时, 该站点将以用户名 "{unique_name}" 注册.')
            if not re.search("^\w+$", unique_name):
                self.log.warning(f"用户名含有除 a-z, A-Z, 0-9, 以及下划线之外的字符, 可能导致注册失败.")
            return unique_name
        else:
            return Monitor.unique_cache[self.client.me]



from datetime import datetime, timedelta
import random
import string

from embykeeper.runinfo import RunStatus

from ..monitor import UniqueUsername
from ..embyboss import EmbybossRegister
from . import BotCheckin


__ignore__ = True


class TemplateCCheckin(BotCheckin):
    bot_use_captcha = False
    unique_cache = UniqueUsername()

    def get_unique_name(self):
        unique_name = self.config.get("unique_name", None)
        if unique_name:
            return unique_name
        else:
            return self.__class__.unique_cache[self.client.me]

    async def start(self):
        random_code = "".join(random.choices(string.ascii_letters + string.digits, k=4))
        if await EmbybossRegister(self.client, self.log, self.get_unique_name(), random_code).run(
            self.bot_username
        ):
            self.log.bind(log=True).info(f"定时开注测试器成功注册机器人 {self.bot_username}.")
            return self.ctx.finish(RunStatus.SUCCESS, "机器人注册成功")
        else:
            interval = self.config.get("interval", 7200)
            self.ctx.next_time = datetime.now() + timedelta(seconds=interval)
            return self.ctx.finish(RunStatus.RESCHEDULE, "机器人未开注")


def use(**kw):
    return type("TemplatedClass", (TemplateCCheckin,), kw)



import json
from typing import Any, List

from loguru import logger

from .utils import CachedFuncProxy
from .config import config


class Cache:
    def __init__(self):
        self._mongo_client = None
        if hasattr(config, "mongodb") and config.mongodb:
            try:
                from pymongo import MongoClient

                self._mongo_client = MongoClient(config.mongodb)
                self._db = self._mongo_client.embykeeper
                self._collection = self._db.cache
            except ImportError:
                logger.warning("没有安装 pymongo 包, 将使用 JSON 存储缓存.")
                self._setup_json_cache()
        else:
            self._setup_json_cache()

    def _setup_json_cache(self):
        self._cache_file = config.basedir / "cache.json"
        self._data = {}
        if self._cache_file.exists():
            try:
                with open(self._cache_file, "r", encoding="utf-8") as f:
                    self._data = json.load(f)
            except json.JSONDecodeError:
                logger.warning("缓存文件损坏, 将使用全新缓存.")

    def get(self, key: str, default: Any = None) -> Any:
        if self._mongo_client:
            result = self._collection.find_one({"_id": key})
            return result["value"] if result else default
        else:
            value = self._data
            try:
                for part in key.split("."):
                    value = value.get(part, {})
                return default if value == {} else value
            except (AttributeError, TypeError):
                return default

    def set(self, key: str, value: Any) -> None:
        if self._mongo_client:
            self._collection.update_one({"_id": key}, {"$set": {"value": value}}, upsert=True)
        else:
            parts = key.split(".")
            current = self._data
            for part in parts[:-1]:
                current = current.setdefault(part, {})
            current[parts[-1]] = value
            with open(self._cache_file, "w", encoding="utf-8") as f:
                json.dump(self._data, f, ensure_ascii=False)

    def delete(self, key: str) -> None:
        if self._mongo_client:
            self._collection.delete_one({"_id": key})
        else:
            parts = key.split(".")
            current = self._data
            path = []

            # 遍历路径，检查每一层
            for part in parts[:-1]:
                if not isinstance(current, dict) or part not in current:
                    return
                current = current[part]
                path.append((part, current))

            # 检查并删除最后一个键
            if isinstance(current, dict) and parts[-1] in current:
                del current[parts[-1]]

                # 清理空字典
                for part, parent in reversed(path):
                    if isinstance(parent, dict) and part in parent and not parent[part]:
                        del parent[part]
                    else:
                        break

            with open(self._cache_file, "w", encoding="utf-8") as f:
                json.dump(self._data, f, ensure_ascii=False, indent=2)

    def find_by_prefix(self, prefix: str) -> List[str]:
        if self._mongo_client:
            return [
                doc["_id"] for doc in self._collection.find({"_id": {"$regex": f"^{prefix}"}}, {"_id": 1})
            ]
        else:

            def get_keys_with_prefix(d, current_path="", keys=None):
                if keys is None:
                    keys = []
                for k, v in d.items():
                    path = f"{current_path}.{k}" if current_path else k
                    if isinstance(v, dict):
                        get_keys_with_prefix(v, path, keys)
                    else:
                        if path.startswith(prefix):
                            keys.append(path)
                return keys

            return get_keys_with_prefix(self._data)

    def delete_by_prefix(self, prefix: str) -> None:
        keys = self.find_by_prefix(prefix)
        for key in keys:
            self.delete(key)

    def delete_many(self, keys: List[str]) -> None:
        """批量删除多个键的缓存

        Args:
            keys: 要删除的键列表
        """
        if self._mongo_client:
            self._collection.delete_many({"_id": {"$in": keys}})
        else:
            # 批量删除所有键，只写入一次文件
            changed = False
            for key in keys:
                parts = key.split(".")
                current = self._data
                path = []

                # 遍历路径，检查每一层
                for part in parts[:-1]:
                    if not isinstance(current, dict) or part not in current:
                        break
                    current = current[part]
                    path.append((part, current))

                # 检查并删除最后一个键
                if isinstance(current, dict) and parts[-1] in current:
                    del current[parts[-1]]
                    changed = True

                    # 清理空字典
                    for part, parent in reversed(path):
                        if isinstance(parent, dict) and part in parent and not parent[part]:
                            del parent[part]
                        else:
                            break

            # 只在有改动时写入一次文件
            if changed:
                with open(self._cache_file, "w", encoding="utf-8") as f:
                    json.dump(self._data, f, ensure_ascii=False, indent=2)


cache: Cache = CachedFuncProxy(lambda: Cache())



import asyncio
from pyrogram.types import Message, InlineKeyboardMarkup
from pyrogram.enums import MessageEntityType
from pyrogram.errors import RPCError

from ..lock import pornemby_messager_enabled, pornemby_alert
from . import Monitor


class PornembyDoubleMonitor(Monitor):
    name = "Pornemby 怪兽自动翻倍"
    chat_user = ["PronembyTGBot2_bot", "PronembyTGBot3_bot", "PornembyBot", "Porn_Emby_Bot"]
    chat_name = "Pornemby"
    chat_keyword = r"击杀者\s+(.*)\s+是否要奖励翻倍"
    additional_auth = ["pornemby_pack"]
    allow_edit = True

    async def on_trigger(self, message: Message, key, reply):
        if pornemby_alert.get(self.client.me.id, False):
            self.log.info(f"由于风险急停不翻倍.")
            return
        for me in message.entities:
            if me.type == MessageEntityType.TEXT_MENTION:
                if me.user.id == self.client.me.id:
                    if isinstance(message.reply_markup, InlineKeyboardMarkup):
                        try:
                            await message.click("🎲开始翻倍游戏")
                        except RPCError:
                            pass
                        else:
                            self.log.info("检测到 Pornemby 怪兽击败, 已点击翻倍.")
                            return

    async def init(self):
        interval = 1
        while True:
            if pornemby_messager_enabled.get(self.client.me.id, False):
                return True
            await asyncio.sleep(interval)
            interval += 1



import logging
import os
from pathlib import Path
import sys
from typing import List
from functools import wraps

import typer
import asyncio
from loguru import logger
from appdirs import user_data_dir

from . import var, __author__, __name__ as __product__, __url__, __version__
from .utils import AsyncTaskPool, show_exception
from .config import config


class AsyncTyper(typer.Typer):
    def async_command(self, *args, **kwargs):
        def decorator(async_func):
            @wraps(async_func)
            def sync_func(*_args, **_kwargs):
                async def main():
                    try:
                        await async_func(*_args, **_kwargs)
                    except typer.Exit as e:
                        sys.exit(e.exit_code)
                    except Exception as e:
                        print("\r", end="", flush=True)
                        logger.critical(f"发生关键错误, {__name__.capitalize()} 将退出.")
                        show_exception(e, regular=False)
                        sys.exit(1)
                    else:
                        logger.info(f"所有任务已完成, 欢迎您再次使用 {__name__.capitalize()}.")

                try:
                    loop = asyncio.new_event_loop()
                    asyncio.set_event_loop(loop)
                    loop.run_until_complete(main())
                except KeyboardInterrupt:
                    print("\r正在停止...\r", end="", flush=True, file=sys.stderr)
                finally:
                    if var.exit_handlers:
                        logger.debug("开始执行退出处理程序.")
                        try:
                            # Wait for exit handlers with timeout
                            loop.run_until_complete(
                                asyncio.wait_for(
                                    asyncio.gather(*[h() for h in var.exit_handlers], return_exceptions=True),
                                    timeout=3,
                                )
                            )
                        except asyncio.TimeoutError:
                            logger.warning("部分退出处理程序超时未完成.")
                        else:
                            logger.debug("退出处理程序执行完成, 开始清理所有任务.")
                    else:
                        logger.debug("未注册退出处理程序, 开始清理所有任务.")

                    # Then cancel remaining tasks
                    tasks = asyncio.all_tasks(loop)
                    for task in tasks:
                        task.cancel()
                    loop.run_until_complete(asyncio.gather(*tasks, return_exceptions=True))
                    loop.run_until_complete(loop.shutdown_asyncgens())
                    print("\r", end="", flush=True)
                    logger.info(f"所有服务已停止并登出, 欢迎您再次使用 {__product__.capitalize()}.")

            self.command(*args, **kwargs)(sync_func)
            return async_func

        return decorator


app = AsyncTyper(
    pretty_exceptions_enable=False,
    rich_markup_mode="rich",
    add_completion=False,
    context_settings={"help_option_names": ["-h", "--help"]},
)


def version(flag):
    if flag:
        print(__version__)
        raise typer.Exit()


def print_example_config(flag):
    if flag:
        print(config.generate_example_config())
        raise typer.Exit()


@app.async_command(
    help=(
        f"欢迎使用 [orange3]{__product__.capitalize()}[/] {__version__} " ":cinema: 无参数默认开启全部功能."
    )
)
async def main(
    config_file: Path = typer.Argument(
        None,
        dir_okay=False,
        allow_dash=True,
        envvar=f"EK_CONFIG_FILE",
        rich_help_panel="参数",
        help="配置文件 (置空以生成)",
    ),
    checkiner: bool = typer.Option(
        False,
        "--checkin",
        "-c",
        rich_help_panel="模块开关",
        help="仅启用 Telegram 签到功能",
    ),
    emby: bool = typer.Option(
        False,
        "--emby",
        "-e",
        rich_help_panel="模块开关",
        help="仅启用 Emby 保活功能",
    ),
    subsonic: bool = typer.Option(
        False,
        "--subsonic",
        "-S",
        rich_help_panel="模块开关",
        help="仅启用 Subsonic 保活功能",
    ),
    monitor: bool = typer.Option(
        False,
        "--monitor",
        "-m",
        rich_help_panel="模块开关",
        help="仅启用群聊监视功能",
    ),
    messager: bool = typer.Option(
        False,
        "--messager",
        "-s",
        rich_help_panel="模块开关",
        help="仅启用自动水群功能",
    ),
    version: bool = typer.Option(
        None,
        "--version",
        "-v",
        rich_help_panel="调试参数",
        callback=version,
        is_eager=True,
        help=f"打印 {__product__.capitalize()} 版本",
    ),
    example_config: bool = typer.Option(
        None,
        "--example-config",
        "-E",
        hidden=True,
        callback=print_example_config,
        is_eager=True,
        help=f"输出范例配置文件",
    ),
    instant: bool = typer.Option(
        False,
        "--instant/--no-instant",
        "-i/-I",
        envvar="EK_INSTANT",
        show_envvar=False,
        rich_help_panel="调试参数",
        help="启动时立刻执行一次任务",
    ),
    once: bool = typer.Option(
        False,
        "--once/--cron",
        "-o/-O",
        rich_help_panel="调试参数",
        help="只执行一次而不进入计划执行模式",
    ),
    verbosity: int = typer.Option(
        False,
        "--debug",
        "-d",
        count=True,
        envvar="EK_DEBUG",
        show_envvar=False,
        rich_help_panel="调试参数",
        help="开启调试模式",
    ),
    debug_cron: bool = typer.Option(
        False,
        envvar="EK_DEBUG_CRON",
        show_envvar=False,
        help="开启任务调试模式, 在三秒后立刻开始执行计划任务",
    ),
    debug_notify: bool = typer.Option(
        False,
        show_envvar=False,
        help="开启日志调试模式, 发送一条日志记录和即时日志记录后退出",
    ),
    simple_log: bool = typer.Option(
        False,
        "--simple-log",
        "-L",
        rich_help_panel="调试参数",
        help="简化日志输出格式",
    ),
    disable_color: bool = typer.Option(
        False,
        "--disable-color",
        "-C",
        rich_help_panel="调试参数",
        help="禁用日志颜色",
    ),
    follow: bool = typer.Option(
        False,
        "--follow",
        "-F",
        rich_help_panel="调试工具",
        help="仅启动消息调试",
    ),
    analyze: bool = typer.Option(
        False,
        "--analyze",
        "-A",
        rich_help_panel="调试工具",
        help="仅启动历史信息分析",
    ),
    dump: List[str] = typer.Option(
        [],
        "--dump",
        "-D",
        rich_help_panel="调试工具",
        help="仅启动更新日志",
    ),
    top: bool = typer.Option(
        False,
        "--top",
        "-T",
        rich_help_panel="调试参数",
        help="执行过程中显示系统状态底栏",
    ),
    play: str = typer.Option(
        None,
        "--play-url",
        "-p",
        rich_help_panel="调试工具",
        help="仅模拟观看一个视频",
    ),
    save: bool = typer.Option(
        False,
        "--save",
        rich_help_panel="调试参数",
        help="记录执行过程中的原始更新日志",
    ),
    public: bool = typer.Option(
        False,
        "--public",
        "-P",
        hidden=True,
        rich_help_panel="调试参数",
        help="启用公共仓库部署模式",
    ),
    windows: bool = typer.Option(
        False,
        "--windows",
        "-W",
        hidden=True,
        rich_help_panel="调试参数",
        help="启用 Windows 安装部署模式",
    ),
    basedir: Path = typer.Option(
        None,
        "--basedir",
        "-B",
        rich_help_panel="调试参数",
        help="设定账号文件的位置",
    ),
    noexit: bool = typer.Option(
        False,
        "--noexit",
        "-N",
        rich_help_panel="调试参数",
        help="要求所有长期任务在没有账号时继续监控等待",
    ),
    clean: bool = typer.Option(
        False,
        "--clean",
        rich_help_panel="调试参数",
        help="显示或清理 Emby 模拟设备和登陆凭据等缓存",
    ),
):
    from .log import initialize, apply_logging_adapter

    var.debug = verbosity
    if verbosity >= 3:
        level = 0
        logging.getLogger("pyrogram.session").setLevel(20)
        logging.getLogger("hpack").setLevel(20)
        asyncio.get_event_loop().set_debug(True)
        apply_logging_adapter(level=10)
    elif verbosity >= 1:
        level = "DEBUG"
    else:
        level = "INFO"

    initialize(level=level, show_path=verbosity and (not simple_log), show_time=not simple_log)
    if disable_color:
        var.console.no_color = True

    msg = " 您可以通过 Ctrl+C 以结束运行." if not public else ""
    logger.info(f"欢迎使用 [orange3]{__product__.capitalize()}[/]! 正在启动, 请稍等.{msg}")
    logger.info(f"当前版本 ({__version__}) 项目页: {__url__}")
    logger.debug(f'命令行参数: "{" ".join(sys.argv[1:])}".')

    basedir = Path(basedir or user_data_dir(__product__))
    basedir.mkdir(parents=True, exist_ok=True)
    if public:
        logger.info(f'工作目录: "{basedir}"')
    else:
        logger.info(f'工作目录: "{basedir}", 您的用户数据相关文件将存储在此处, 请妥善保管.')
        docker = bool(os.environ.get("EK_IN_DOCKER", False))
        if docker:
            logger.info("当前在 Docker 容器中运行, 请确认该目录已挂载, 否则文件将在容器重建后丢失.")
    if verbosity:
        logger.warning(f"您当前处于调试模式: 日志等级 {verbosity}.")
        app.pretty_exceptions_enable = True

    config.basedir = basedir
    config.windows = windows
    config.public = public

    if public:
        from .public import public_entrypoint

        if not await public_entrypoint():
            raise typer.Exit(1)
    else:
        if not await config.reload_conf(config_file):
            raise typer.Exit(1)

    if verbosity >= 2:
        config.nofail = False
    if not config.nofail:
        logger.warning(f"您当前处于调试模式: 错误将会导致程序停止运行.")
    if debug_cron:
        config.debug_cron = True
        logger.warning("您当前处于计划任务调试模式, 将在 10 秒后运行计划任务.")
    config.noexit = noexit

    if not checkiner and not monitor and not emby and not messager and not subsonic:
        checkiner = True
        emby = True
        subsonic = True
        monitor = True
        messager = True

    if config.mongodb:
        logger.info(f"正在连接到 MongoDB 缓存, 请稍候.")
        try:
            from .cache import cache

            cache.set("test", "test")
            assert cache.get("test", None) == "test"
            cache.delete("test")
        except Exception as e:
            logger.error(f"MongoDB 缓存连接失败: {e}, 程序将退出.")
            show_exception(e, regular=False)
            return

    if clean:
        from .clean import cleaner

        return await cleaner()

    if follow:
        from .telegram.debug import follower

        return await follower()

    if top:
        from .topper import topper

        if not (var.console.is_terminal and var.console.is_interactive):
            logger.warning("在非交互模式下启用底栏可能会导致显示异常.")
        asyncio.create_task(topper())

    if play:
        from .emby.main import EmbyManager

        return await EmbyManager().play_url(play)

    if save:
        from .telegram.debug import saver

        asyncio.create_task(saver())

    if analyze:
        from .telegram.debug import analyzer

        indent = " " * 23
        chats = typer.prompt(indent + "请输入群组用户名 (以空格分隔)").split()
        keywords = typer.prompt(indent + "请输入关键词 (以空格分隔)", default="", show_default=False)
        keywords = keywords.split() if keywords else []
        timerange = typer.prompt(indent + '请输入时间范围 (以"-"分割)', default="", show_default=False)
        timerange = timerange.split("-") if timerange else []
        limit = typer.prompt(indent + "请输入各群组最大获取数量", default=10000, type=int)
        outputs = typer.prompt(indent + "请输入最大输出数量", default=1000, type=int)
        return await analyzer(chats, keywords, timerange, limit, outputs)

    if dump:
        from .telegram.debug import dumper

        return await dumper(dump)

    if debug_notify:
        from .telegram.debug import debug_notifier

        return await debug_notifier()

    try:
        checkin_man = None
        if checkiner:
            from .telegram.checkin_main import CheckinerManager

            checkin_man = CheckinerManager()

        monitor_man = None
        if monitor:
            from .telegram.monitor_main import MonitorManager

            monitor_man = MonitorManager()

        message_man = None
        if messager:
            from .telegram.message_main import MessageManager

            message_man = MessageManager()

        emby_man = None
        if emby:
            from .emby.main import EmbyManager

            emby_man = EmbyManager()

        subsonic_man = None
        if subsonic:
            from .subsonic.main import SubsonicManager

            subsonic_man = SubsonicManager()

        pool = AsyncTaskPool()
        if instant and not debug_cron:
            if checkin_man:
                pool.add(checkin_man.run_all(instant=True), "站点签到")
            if emby_man:
                pool.add(emby_man.run_all(instant=True), "Emby 保活")
            if subsonic_man:
                pool.add(subsonic_man.run_all(instant=True), "Subsonic 保活")
            await pool.wait()
            logger.debug("启动时立刻执行签到和保活: 已完成.")
        streams = None
        if config.notifier.enabled:
            from .telegram.notify import start_notifier

            streams = await start_notifier()
        if not once:
            if checkin_man:
                pool.add(checkin_man.schedule_all(), "站点签到")
            if monitor_man:
                pool.add(monitor_man.run_all(), "群组监控")
            if message_man:
                pool.add(message_man.run_all(), "自动水群")
            if emby_man:
                pool.add(emby_man.schedule_all(), "Emby 保活")
            if subsonic_man:
                pool.add(subsonic_man.schedule_all(), "Subsonic 保活")
        if config.noexit:
            logger.info("处于长期监控模式, 当没有账号时将继续监控等待新配置.")
            pool.add(asyncio.Event().wait(), "账号配置文件监控")
        try:
            async for t in pool.as_completed():
                try:
                    await t
                except asyncio.CancelledError:
                    logger.debug(f"任务 {t.get_name()} 被取消.")
                except Exception as e:
                    logger.debug(f"任务 {t.get_name()} 出现错误, 模块可能停止运行.")
                    show_exception(e, regular=False)
                    if not config.nofail:
                        raise
                else:
                    logger.debug(f"任务 {t.get_name()} 成功结束.")
        finally:
            if streams:
                await asyncio.gather(*[stream.join() for stream in streams])
    finally:
        from .runinfo import RunContext

        RunContext.cancel_all()


if __name__ == "__main__":
    app()



from . import BotCheckin

__ignore__ = True


class ZhushouSGKCheckin(BotCheckin):
    name = "助手社工库"
    bot_username = "sgk001_bot"
    bot_checkin_cmd = "/checkin"
    bot_checked_keywords = ["今日已签到"]



import asyncio
from datetime import datetime
import random
import string
from urllib.parse import quote
import uuid
from typing import Iterable, List, Union, Optional
import re

from loguru import logger
from curl_cffi.requests import AsyncSession, Response, RequestsError
from pydantic import BaseModel, ValidationError

from embykeeper import __version__
from embykeeper.utils import get_proxy_str, show_exception, truncate_str
from embykeeper.cache import cache
from embykeeper.schema import EmbyAccount
from embykeeper.config import config

logger = logger.bind(scheme="embywatcher")


class EmbyError(Exception):
    pass


class EmbyRequestError(EmbyError):
    pass


class EmbyConnectError(EmbyError):
    pass


class EmbyLoginError(EmbyRequestError):
    pass


class EmbyStatusError(EmbyRequestError):
    pass


class EmbyPlayError(EmbyError):
    pass


class EmbyEnv(BaseModel):
    client: str
    device: str
    device_id: str
    client_version: str
    useragent: str


class Emby:
    playing_count = 0

    def __init__(self, account: EmbyAccount):
        self.a = account

        self._env = None
        self._token = None
        self._user_id = None

        self.run_id = str(uuid.uuid4()).upper()
        self.cf_clearance = None
        self.useragent = None
        self.items = {}

        self.log = logger.bind(server=self.a.name or self.hostname, username=self.a.username)

    @property
    def proxy(self):
        return config.proxy if self.a.use_proxy else None

    @property
    def hostname(self):
        return self.a.url.host

    @property
    def token(self):
        if not self._token:
            self._load_credentials()
        return self._token

    @property
    def env(self):
        if not self._env:
            self._load_env()
        if not self._env:
            self._env = self.get_fake_env()
        return self._env

    @property
    def user_id(self):
        if not self._user_id:
            self._load_credentials()
        return self._user_id

    def _load_credentials(self):
        data: dict = cache.get(f"emby.credential.{self.hostname}.{self.a.username}", {})
        self._token = data.get("token", None)
        self._user_id = data.get("userid", None)

    def _load_env(self):
        cache_key = f"emby.env.{self.hostname}.{self.a.username}"
        data: dict = cache.get(cache_key, {})
        if data:
            # 检查用户配置是否与缓存一致
            should_clear = False
            for key, user_value in {
                "client_version": self.a.client_version,
                "client": self.a.client,
                "device": self.a.device,
                "device_id": self.a.device_id,
                "useragent": self.a.useragent,
            }.items():
                if user_value and data.get(key) != user_value:
                    should_clear = True
                    break

            if should_clear:
                logger.info("账户设置已修改, 将重新生成环境 (Headers).")
                self._env = None
                cache.delete(cache_key)
            else:
                try:
                    self._env = EmbyEnv.model_validate(data)
                except ValidationError:
                    logger.warning("缓存加载失败, 将重新生成环境 (Headers).")
                    self._env = None

    @staticmethod
    def get_random_device():
        from faker import Faker

        device_type = random.choice(("iPhone", "iPad"))

        # All patterns with their weights
        patterns = [
            ("chinese_normal", 20),
            ("chinese_lastname_pinyin", 40),
            ("chinese_firstname_pinyin", 10),
            ("english_normal", 20),
            ("english_upper", 10),
            ("english_name_only", 10),
        ]

        pattern = random.choices([p[0] for p in patterns], weights=[p[1] for p in patterns])[0]

        if pattern.startswith("chinese"):
            fake = Faker("zh_CN")
            surname = fake.last_name()
            given_name = fake.first_name_male() if random.random() < 0.5 else fake.first_name_female()

            if pattern == "chinese_normal":
                return f"{surname}{given_name}的{device_type}"
            else:
                from xpinyin import Pinyin

                p = Pinyin()
                if pattern == "chinese_lastname_pinyin":
                    pinyin = p.get_pinyin(surname).capitalize()
                    return f"{pinyin}的{device_type}"
                else:  # chinese_firstname_pinyin
                    pinyin = "".join([word[0].upper() for word in p.get_pinyin(given_name).split("-")])
                    return f"{pinyin}的{device_type}"
        else:
            fake = Faker("en_US")
            name = fake.first_name()

            if pattern == "english_normal":
                return f"{name}'s {device_type}"
            elif pattern == "english_upper":
                return f"{name.upper()}{device_type.upper()}"
            else:  # english_name_only
                return name

    @staticmethod
    def get_device_uuid():
        rd = random.Random()
        rd.seed(uuid.getnode())
        return uuid.UUID(int=rd.getrandbits(128))

    def get_fake_env(self):
        cached_env: dict = cache.get(f"emby.env.{self.hostname}.{self.a.username}", {})

        # 按优先级获取各个值
        is_filebar = random.random() < 0.2
        version = (
            self.a.client_version
            or cached_env.get("client_version")
            or f"1.3.{random.randint(34, 34) if is_filebar else random.randint(16, 30)}"
        )
        client = self.a.client or cached_env.get("client") or ("Filebar" if is_filebar else "Fileball")
        device = self.a.device or cached_env.get("device") or self.get_random_device()
        device_id = self.a.device_id or cached_env.get("device_id") or str(uuid.uuid4()).upper()
        useragent = self.useragent or self.a.useragent or cached_env.get("ua") or f"{client}/{version}"

        data = {
            "client": client,
            "device": device,
            "device_id": device_id,
            "client_version": version,
            "useragent": useragent,
        }

        env = EmbyEnv(**data)
        cache.set(f"emby.env.{self.hostname}.{self.a.username}", data)
        return env

    def build_headers(self):
        headers = {}
        auth_headers = {
            "Client": self.env.client,
            "Device": self.env.device,
            "DeviceId": self.env.device_id,
            "Version": self.env.client_version,
        }
        auth_header = ",".join([f"{k}={quote(str(v))}" for k, v in auth_headers.items()])
        full_auth_header = f'MediaBrowser Token={self.token or ""},Emby UserId={self.run_id},{auth_header}'
        headers["User-Agent"] = self.useragent or self.env.useragent
        headers["Accept-Language"] = "zh-CN,zh-Hans;q=0.9"
        headers["Content-Type"] = "application/json"
        headers["Accept"] = "*/*"
        headers["X-Emby-Authorization"] = full_auth_header
        if self.token:
            headers["X-Emby-Token"] = self.token
        return headers

    def _get_session(self) -> AsyncSession:
        cookies = {}
        if self.cf_clearance:
            cookies["cf_clearance"] = self.cf_clearance

        return AsyncSession(
            verify=False,
            headers=self.build_headers(),
            cookies=cookies,
            proxy=get_proxy_str(self.proxy),
            timeout=10.0,
            impersonate="chrome",
            allow_redirects=True,
            default_headers=False,
        )

    async def _request(self, method: str, path: str, _login=False, **kw) -> Response:

        if path.startswith(("http://", "https://")):
            url = path
        else:
            base_url = f"{self.a.url.scheme}://{self.a.url.host}:{self.a.url.port}"
            url = f"{base_url}/{path.lstrip('/')}"

        last_err = None
        for _ in range(3):
            try:
                async with self._get_session() as session:
                    resp: Response = await session.request(method, url, **kw)
                    if resp.status_code == 401 and self.a.username and not _login:
                        if not await self.login():
                            raise EmbyLoginError("无法登陆到服务器")
                        continue
                    elif resp.status_code in (502, 503, 504):
                        await asyncio.sleep(random.random() * 2 + 0.5)
                        continue
                    elif resp.status_code == 403 and (
                        "cf-wrapper" in resp.text or "Just a moment" in resp.text
                    ):
                        if self.cf_clearance:
                            raise EmbyStatusError("访问失败: Cloudflare 验证码解析后依然有验证")
                        await self.use_cfsolver()
                        continue
                    elif not resp.ok and not _login:
                        raise EmbyStatusError(f"访问失败: 异常 HTTP 代码 {resp.status_code} (URL = {url})")
                    else:
                        return resp
            except RequestsError as e:
                last_err = e
                await asyncio.sleep(random.random() + 0.5)

        error_msg = re.sub(r"\s+See\s+.*?\s+first for more details\.\.?", "", str(last_err))
        raise EmbyConnectError(f"{last_err.__class__.__name__}: {error_msg}")

    async def use_cfsolver(self):
        from embykeeper.cloudflare import get_cf_clearance

        if not self.a.cf_challenge:
            if self.proxy:
                self.log.warning(
                    f"该站点已启用 Cloudflare 保护, 请尝试浏览器以同样的代理访问: {self.a.url}"
                    "以解除 Cloudflare IP 限制, 然后再次运行.\n"
                    '或者, 高级用户可以使用 "cf_challenge = true" 配置项以允许尝试解析验证码.'
                )
            else:
                self.log.warning(
                    f'该站点已启用 Cloudflare 保护, 请使用 "cf_challenge = true" 配置项以允许尝试解析验证码.'
                )
        self.log.info(f"该站点已启用 Cloudflare 保护, 即将请求解析.")
        if self.proxy:
            if self.proxy.scheme != "socks5":
                self.log.warning(
                    f"该站点验证解析仅支持 SOCKS5 代理，由于当前代理协议不支持, 将尝试不使用代理."
                )
                self.a.use_proxy = False
            else:
                self.log.info(
                    f"验证码解析将使用代理, 可能导致解析失败, 若失败请使用"
                    '"use_proxy = false" 以禁用该站点的代理.'
                )
        try:
            cf_clearance, useragent = await get_cf_clearance(self.a.url, self.proxy)
            if not cf_clearance:
                self.log.warning(f"Cloudflare 验证码解析失败.")
                return False
            else:
                self.cf_clearance = cf_clearance
                self.useragent = useragent
                return True
        except Exception as e:
            self.log.warning(f"Cloudflare 验证码解析时出现错误.")
            show_exception(e, regular=False)
            return False

    async def login(self) -> dict:
        """Login to Emby server and get authentication token."""

        if self.a.username is None or self.a.password is None:
            self.log.warning("没有提供用户名或密码, 无法登陆, 执行失败.")
            return None

        data = {
            "Username": self.a.username,
            "Pw": self.a.password,
        }

        resp = await self._request(
            "POST",
            "/Users/AuthenticateByName",
            json=data,
            _login=True,
        )

        if resp.status_code == 401:
            self.log.warning(f"用户名或密码错误, 执行失败.")
            return None

        if resp.status_code != 200:
            self.log.warning(f"登陆时出现错误 ({resp.status_code}), 执行失败.")
            return None

        user: dict = resp.json()
        self._token = user.get("AccessToken", None)
        self._user_id = user.get("User", {}).get("Id")
        if self.token and self.user_id:
            cache_data = {
                "token": self.token,
                "userid": self.user_id,
            }
            cache.set(f"emby.credential.{self.hostname}.{self.a.username}", cache_data)
            return self.token

    async def play(self, item: Union[dict, int], time: float = 10):
        if isinstance(item, dict):
            try:
                iid = item["Id"]
                iname = item["Name"]
            except KeyError:
                raise EmbyPlayError("无法解析视频信息")
        else:
            iid = item
            iname = "(请求播放的视频)"

        playback_info_data = {
            "DeviceProfile": {
                "CodecProfiles": [
                    {
                        "ApplyConditions": [
                            {
                                "IsRequired": False,
                                "Value": "true",
                                "Condition": "NotEquals",
                                "Property": "IsAnamorphic",
                            },
                            {
                                "IsRequired": False,
                                "Value": "high|main|baseline|constrained baseline",
                                "Condition": "EqualsAny",
                                "Property": "VideoProfile",
                            },
                            {
                                "IsRequired": False,
                                "Value": "80",
                                "Condition": "LessThanEqual",
                                "Property": "VideoLevel",
                            },
                            {
                                "IsRequired": False,
                                "Value": "true",
                                "Condition": "NotEquals",
                                "Property": "IsInterlaced",
                            },
                        ],
                        "Type": "Video",
                        "Codec": "h264",
                    },
                    {
                        "ApplyConditions": [
                            {
                                "IsRequired": False,
                                "Value": "true",
                                "Condition": "NotEquals",
                                "Property": "IsAnamorphic",
                            },
                            {
                                "IsRequired": False,
                                "Value": "high|main|main 10",
                                "Condition": "EqualsAny",
                                "Property": "VideoProfile",
                            },
                            {
                                "IsRequired": False,
                                "Value": "175",
                                "Condition": "LessThanEqual",
                                "Property": "VideoLevel",
                            },
                            {
                                "IsRequired": False,
                                "Value": "true",
                                "Condition": "NotEquals",
                                "Property": "IsInterlaced",
                            },
                        ],
                        "Type": "Video",
                        "Codec": "hevc",
                    },
                ],
                "SubtitleProfiles": [
                    {"Method": "Embed", "Format": "ass"},
                    {"Method": "Embed", "Format": "ssa"},
                    {"Method": "Embed", "Format": "subrip"},
                    {"Method": "Embed", "Format": "sub"},
                    {"Method": "Embed", "Format": "pgssub"},
                    {"Method": "External", "Format": "subrip"},
                    {"Method": "External", "Format": "sub"},
                    {"Method": "External", "Format": "ass"},
                    {"Method": "External", "Format": "ssa"},
                    {"Method": "External", "Format": "vtt"},
                    {"Method": "External", "Format": "ass"},
                    {"Method": "External", "Format": "ssa"},
                ],
                "MaxStreamingBitrate": 40000000,
                "DirectPlayProfiles": [
                    {
                        "Container": "mov,mp4,mkv,webm",
                        "Type": "Video",
                        "VideoCodec": "h264,hevc,dvhe,dvh1,h264,hevc,hev1,mpeg4,vp9",
                        "AudioCodec": "aac,mp3,wav,ac3,eac3,flac,truehd,dts,dca,opus",
                    }
                ],
                "TranscodingProfiles": [
                    {
                        "MinSegments": 2,
                        "AudioCodec": "aac,mp3,wav,ac3,eac3,flac,opus",
                        "VideoCodec": "hevc,h264,mpeg4",
                        "BreakOnNonKeyFrames": True,
                        "Type": "Video",
                        "Protocol": "hls",
                        "MaxAudioChannels": "6",
                        "Container": "ts",
                        "Context": "Streaming",
                    }
                ],
                "ContainerProfiles": [],
                "MusicStreamingTranscodingBitrate": 40000000,
                "ResponseProfiles": [{"MimeType": "video\/mp4", "Container": "m4v", "Type": "Video"}],
                "MaxStaticBitrate": 40000000,
            }
        }

        resp = await self._request(
            method="GET",
            path=f"/Videos/{iid}/AdditionalParts",
            params=dict(
                Fields="PrimaryImageAspectRatio,UserData,CanDelete",
                IncludeItemTypes="Playlist,BoxSet",
                Recursive=True,
                SortBy="SortName",
            ),
        )

        resp = await self._request(
            method="POST",
            path=f"/Items/{iid}/PlaybackInfo",
            params=dict(
                AutoOpenLiveStream=False,
                IsPlayback=False,
                MaxStreamingBitrate=40000000,
                StartTimeTicks=0,
                UserID=self.user_id,
            ),
            json=playback_info_data,
        )
        playback_info = resp.json()

        play_session_id = playback_info.get("PlaySessionId", "")
        if "MediaSources" in playback_info:
            media_source_id = playback_info["MediaSources"][0]["Id"]
            direct_stream_url = playback_info["MediaSources"][0].get("DirectStreamUrl", None)
        else:
            media_source_id = "".join(
                random.choice(string.ascii_lowercase + string.digits) for _ in range(32)
            )
            direct_stream_url = None

        await asyncio.sleep(random.uniform(1, 3))

        # 模拟播放
        for i in range(4):
            if i:
                IsPlayback = True
                AutoOpenLiveStream = True
            else:
                IsPlayback = False
                AutoOpenLiveStream = False

            resp = await self._request(
                method="POST",
                path=f"/Items/{iid}/PlaybackInfo",
                params=dict(
                    AudioStreamIndex=1,
                    AutoOpenLiveStream=AutoOpenLiveStream,
                    IsPlayback=IsPlayback,
                    MaxStreamingBitrate=42000000,
                    MediaSourceId=str(media_source_id),
                    StartTimeTicks=0,
                    UserID=self.user_id,
                ),
                json=playback_info_data,
            )

        def get_playing_data(tick, update=False, stop=False):
            data = {
                "SubtitleOffset": 0,
                "MaxStreamingBitrate": 420000000,
                "MediaSourceId": str(media_source_id),
                "SubtitleStreamIndex": -1,
                "VolumeLevel": 100,
                "PlaybackRate": 1,
                "PlaybackStartTimeTicks": int(datetime.now().timestamp() // 10 * 10 * 10000000),
                "PositionTicks": tick,
                "PlaySessionId": play_session_id,
            }
            if update:
                data["EventName"] = "timeupdate"
            if stop:
                queue = []
            else:
                queue = [{"Id": str(iid), "PlaylistItemId": "playlistItem0"}]
            data.update(
                {
                    "PlaylistLength": 1,
                    "NowPlayingQueue": queue,
                    "IsMuted": False,
                    "PlaylistIndex": 0,
                    "ItemId": str(iid),
                    "RepeatMode": "RepeatNone",
                    "AudioStreamIndex": -1,
                    "PlayMethod": "DirectStream",
                    "CanSeek": True,
                    "IsPaused": False,
                }
            )
            return data

        async def stream():
            url = direct_stream_url or f"/Videos/{iid}/stream"
            length = 0
            last_err_time = datetime.now()
            while True:
                resp = await self._request(
                    method="GET",
                    path=url,
                    stream=True,
                    max_recv_speed=1024,
                    timeout=None,
                    headers={
                        "Range": f"bytes={length}-",
                        "User-Agent": "VLC/3.0.21 LibVLC/3.0.21",
                        "X-Playback-Session-Id": play_session_id,
                    },
                )
                try:
                    async for i in resp.aiter_content(chunk_size=1024):
                        length += len(i)
                        del i
                        await asyncio.sleep(random.random())
                        if random.random() < 0.01:
                            continue
                except RequestsError:
                    if (datetime.now() - last_err_time).total_seconds() > 5:
                        self.log.debug("流媒体文件访问错误, 正在重试.")
                        last_err_time = datetime.now()
                        continue
                    else:
                        raise
                finally:
                    await resp.aclose()

        stream_task = asyncio.create_task(stream())
        rt = random.uniform(5, 10)
        self.log.info(f'开始模拟加载视频 "{truncate_str(iname, 10)}" ({rt:.0f} 秒).')
        await asyncio.sleep(rt)
        self.log.info(f'开始发送视频 "{truncate_str(iname, 10)}" 发送进度.')
        Emby.playing_count += 1
        try:
            await asyncio.sleep(random.uniform(1, 3))
            try:
                resp = await self._request(
                    method="POST",
                    path="/Sessions/Playing",
                    json=get_playing_data(0),
                )
            except EmbyRequestError as e:
                raise EmbyPlayError(f"无法开始播放: {e}")
            t = time

            last_report_t = t
            progress_errors = 0
            report_interval = 5  # Start with 5 seconds
            report_count = 0
            max_interval = 300  # 5 minutes in seconds
            while t > 0:
                if progress_errors > 12:
                    raise EmbyPlayError("播放状态设定错误次数过多")
                if last_report_t and last_report_t - t > report_interval:
                    self.log.info(f'正在播放: "{truncate_str(iname, 10)}" (还剩 {t:.0f} 秒).')
                    last_report_t = t
                    report_count += 1
                    # After 3 reports at current interval, double the interval
                    if report_count >= 3:
                        report_count = 0
                        report_interval = min(report_interval * 2, max_interval)
                st = min(10, t)
                await asyncio.sleep(st)
                t -= st
                tick = int((time - t) * 10000000)
                payload = get_playing_data(tick, update=True)
                try:
                    resp = await asyncio.wait_for(
                        self._request(
                            method="POST",
                            path="/Sessions/Playing/Progress",
                            json=payload,
                        ),
                        10,
                    )
                except Exception as e:
                    self.log.debug(f"播放状态设定错误: {e}")
                    progress_errors += 1
            await asyncio.sleep(random.uniform(1, 3))
        finally:
            Emby.playing_count -= 1
            stream_task.cancel()
            try:
                await stream_task
            except asyncio.CancelledError:
                pass
            except Exception as e:
                self.log.warning(f"模拟播放时, 访问流媒体文件失败.")
                show_exception(e)

        try:
            final_percentage = random.uniform(0.95, 1.0)
            final_tick = int((time * final_percentage) // 10 * 10 * 10000000)
            await self._request(
                method="POST",
                path="/Sessions/Playing/Progress",
                json=get_playing_data(final_tick, stop=True),
            )
            self.log.info(f"播放完成, 共 {time:.0f} 秒.")
            return True
        except Exception as e:
            raise EmbyPlayError(f"由于连接错误或服务器错误无法停止播放: {e}")

    async def load_main_page(self):
        views = await self._request(
            method="GET",
            path=f"/Users/{self.user_id}/Views",
            params=dict(IncludeExternalContent=False),
        )

        col_ids = []
        for i in views.json().get("Items", []):
            cid: str = i.get("Id", None)
            type: str = i.get("CollectionType")
            if cid and type and type.lower() in ("movies", "tvshows"):
                col_ids.append(cid)
        await asyncio.sleep(random.uniform(0.1, 0.3))

        user = await self._request(method="GET", path=f"/Users/{self.user_id}")
        last_login_date = user.json().get("LastLoginDate", None)
        await asyncio.sleep(random.uniform(0.1, 0.3))

        await self._request(
            method="GET",
            path=f"/DisplayPreferences/usersettings",
            params=dict(client="emby", userId=self.user_id),
        )
        await asyncio.sleep(random.uniform(0.1, 0.3))

        await self.get_resume_items(media_types=["Video"])
        await asyncio.sleep(random.uniform(0.1, 0.3))
        await self.get_resume_items(media_types=["Audio"])
        await asyncio.sleep(random.uniform(0.1, 0.3))

        for cid in col_ids[:25]:
            items = await self.get_latest_items(parent_id=cid)
            for item in items:
                try:
                    iid = item["Id"]
                    self.items[iid] = item
                except KeyError:
                    pass

        if not self.items:
            if col_ids:
                self.log.info("无法获取最新视频, 尝试从文件夹中读取.")

                for col_id in col_ids[:3]:
                    await asyncio.sleep(4)
                    items = await self.get_folder_items(parent_id=col_id)
                    for item in items:
                        try:
                            iid = item["Id"]
                            self.items[iid] = item
                        except KeyError:
                            pass
                    if len(self.items) >= 3:
                        break

        return last_login_date

    async def get_latest_items(
        self,
        enable_image_types=None,
        fields=None,
        limit=16,
        group_items=True,
        parent_id=None,
        **kw,
    ) -> List[dict]:
        if not enable_image_types:
            enable_image_types = ["Primary", "Backdrop", "Thumb"]
        if not fields:
            fields = [
                "PrimaryImageAspectRatio",
                "BasicSyncInfo",
                "ProductionYear",
                "Status",
                "EndDate",
                "CanDelete",
            ]
        resp = await self._request(
            method="GET",
            path=f"/Users/{self.user_id}/Items/Latest",
            params={
                "EnableImageTypes": ",".join(enable_image_types),
                "Fields": ",".join(fields),
                "GroupItems": group_items,
                "Limit": limit,
                "ParentId": parent_id,
                **kw,
            },
        )
        return resp.json()

    async def get_resume_items(
        self,
        enable_image_types=None,
        fields=None,
        limit=12,
        media_types=None,
        **kw,
    ) -> List[dict]:
        if not enable_image_types:
            enable_image_types = ["Primary", "Backdrop", "Thumb"]
        if not fields:
            fields = ["PrimaryImageAspectRatio", "BasicSyncInfo", "ProductionYear", "CanDelete"]
        if not media_types:
            media_types = ["Video"]
        resp = await self._request(
            method="GET",
            path=f"/Users/{self.user_id}/Items/Resume",
            params={
                "EnableImageTypes": ",".join(enable_image_types),
                "Fields": ",".join(fields),
                "Limit": limit,
                "MediaTypes": ",".join(media_types),
                "Recursive": "true",
                **kw,
            },
        )
        return resp.json()

    async def get_folder_items(
        self,
        parent_id,
        enable_image_types=None,
        fields=None,
        limit=50,
        **kw,
    ) -> List[dict]:
        if not enable_image_types:
            enable_image_types = ["Primary", "Backdrop", "Thumb"]
        if not fields:
            fields = ["BasicSyncInfo", "CanDelete", "PrimaryImageAspectRatio", "ProductionYear"]
        resp = await self._request(
            method="GET",
            path=f"/Users/{self.user_id}/Items",
            params={
                "EnableImageTypes": ",".join(enable_image_types),
                "Fields": ",".join(fields),
                "ImageTypeLimit": 1,
                "IncludeItemTypes": "Movie",
                "Limit": limit,
                "ParentId": parent_id,
                "Recursive": "true",
                "SortBy": "SortName",
                "SortOrder": "Ascending",
                "StartIndex": 0,
                **kw,
            },
        )
        return resp.json().get("Items", [])

    async def get_item(self, iid, **kw) -> dict:
        resp = await self._request(method="GET", path=f"/Users/{self.user_id}/Items/{iid}")
        return resp.json()

    async def get_user(self) -> dict:
        """Get current user information."""
        response = await self._request("GET", f"/Users/{self.user_id}")
        return response.json()

    async def mark_played(self, item_id: str) -> bool:
        """Mark an item as played."""
        response = await self._request("POST", f"/Users/{self.user_id}/PlayedItems/{item_id}")
        return response.status_code == 200

    async def watch(self):
        """Play one or more videos until account time requirement played."""

        if self.a.continuous:
            self.log.info("开始连续播放视频")
        else:
            try:
                if isinstance(self.a.time, Iterable):
                    req_time = random.uniform(*self.a.time)
                else:
                    req_time = self.a.time
            except TypeError:
                self.log.warning(
                    f"无法解析 time 配置, 请检查配置: {self.a.time} (应该为数字或两个数字的数组)."
                )
                return False
            msg = " (允许播放多个)" if self.a.allow_multiple else ""
            msg = f"开始播放视频{msg}, 共需播放 {req_time:.0f} 秒."
            self.log.info(msg)

        played_time = 0
        last_played_time = 0
        played_videos = 0
        retry = 0
        failed_items = []
        failed_reasons = {"invalid": 0, "no_length": 0, "wrong_type": 0, "short_length": 0}

        while True:
            shuffled_items = list(self.items.items())
            random.shuffle(shuffled_items)

            for iid, item in shuffled_items:
                try:
                    if iid in failed_items:
                        failed_reasons["invalid"] += 1
                        continue
                except KeyError:
                    continue
                media_type = item.get("MediaType", None)
                if not media_type == "Video":
                    failed_reasons["wrong_type"] += 1
                    continue
                total_ticks = item.get("RunTimeTicks", None)
                if not total_ticks:
                    if self.a.allow_stream:
                        total_ticks = min(req_time, random.randint(480, 720)) * 10000000
                    else:
                        failed_reasons["no_length"] += 1
                        continue
                total_time = total_ticks / 10000000
                if req_time - played_time > total_time:
                    if (not self.a.allow_multiple) and (not self.a.continuous):
                        failed_reasons["short_length"] += 1
                        failed_items.append(iid)
                        continue
                    play_time = total_time
                else:
                    play_time = max(req_time - played_time, 10)
                name = truncate_str(item.get("Name", "(未命名视频)"), 10)
                self.log.info(f'开始播放 "{name}" ({play_time:.0f} 秒).')
                while True:
                    try:
                        await self.play(item, time=play_time)
                        await asyncio.sleep(random.random())
                        item = await self.get_item(iid)
                        play_count = item.get("UserData", {}).get("PlayCount", 0)
                        if play_count < 1:
                            raise EmbyPlayError("播放后播放数低于 1")
                        self.log.info(f"[yellow]成功播放视频[/], 当前该视频播放 {play_count} 次.")
                        played_videos += 1
                        played_time += play_time
                        if played_time >= req_time - 1:
                            self.log.bind(log=True).info(f"保活成功, 共播放 {played_videos} 个视频.")
                            return True
                        else:
                            self.log.info(f"还需播放 {req_time - played_time:.0f} 秒.")
                            rt = random.uniform(5, 15)
                            self.log.info(f"等待 {rt:.0f} 秒后播放下一个.")
                            await asyncio.sleep(rt)
                            break
                    except EmbyError as e:
                        retry += 1
                        if retry > config.emby.retries:
                            self.log.warning(f"超过最大重试次数, 保活失败: {e}.")
                            return False
                        else:
                            rt = random.uniform(30, 60)
                            if isinstance(e, EmbyPlayError):
                                self.log.info(f"播放错误, 等待 {rt:.0f} 秒后重试: {e}.")
                            else:
                                self.log.info(f"连接失败, 等待 {rt:.0f} 秒后重试: {e}.")
                            await asyncio.sleep(rt)
                    except Exception as e:
                        self.log.warning(f"发生错误, 保活失败.")
                        show_exception(e, regular=False)
                        return False
            else:
                if len(failed_items) == len(self.items):
                    reasons = []
                    if failed_reasons["invalid"]:
                        reasons.append(f"{failed_reasons['invalid']} 个视频信息无效")
                    if failed_reasons["no_length"]:
                        reasons.append(f"{failed_reasons['no_length']} 个视频无法获取时长")
                    if failed_reasons["wrong_type"]:
                        reasons.append(f"{failed_reasons['wrong_type']} 个非视频项目")
                    if failed_reasons["short_length"]:
                        reasons.append(
                            f"{failed_reasons['short_length']} 个视频时长不足 (未开启 allow_multiple)"
                        )
                    self.log.warning(f"所有视频均不符合要求, 保活失败. 其中: {', '.join(reasons)}")
                elif played_time > last_played_time:
                    last_played_time = played_time
                    continue
                else:
                    self.log.warning(f"由于没有成功播放视频, 保活失败, 请重新检查配置.")
                    return False

    @staticmethod
    def parse_date(date_str: str) -> Optional[datetime]:
        if not date_str:
            return None
        try:
            return datetime.fromisoformat(date_str.replace("Z", "+00:00"))
        except (ValueError, AttributeError):
            return None



from . import BotCheckin

__ignore__ = True


class AgentSGKCheckin(BotCheckin):
    name = "007 社工库"
    bot_username = "sgk007_bot"
    bot_checkin_cmd = "/sign"
    bot_checked_keywords = ["今日已签到"]
